(***********************************************************************)
(*                                                                     *)
(*                                Sklml                                *)
(*                                                                     *)
(*                 The OCaml parallel skeleton system                  *)
(*                                                                     *)
(*               EPI Pomdapi, INRIA Paris - Rocquencourt               *)
(*                                                                     *)
(*  Copyright 2007-2012 INRIA.                                         *)
(*  Institut National de Recherche en Informatique et en Automatique.  *)
(*  All rights reserved. This file is distributed only by permission.  *)
(*                                                                     *)
(*  Quentin Carbonneaux <Quentin.Carbonneaux@inria.fr>                 *)
(*  Francois Clement <Francois.Clement@inria.fr>                       *)
(*  Pierre Weis <Pierre.Weis@inria.fr>                                 *)
(*                                                                     *)
(***********************************************************************)

(* $Id: main.ml,v 1.33 2013/03/29 13:21:41 fclement Exp $ *)

(** This generates primes using sklml and C helper code *)

open Skl;;

let lift_set_config k conf = k (Options.set_config conf);;

let stop_cond =
  Expor.register
    (lift_set_config
       (fun () ->
         Sklstream.map (Pgen.stop_vcond (Options.get_pool_size ()))))
;;

let next_perms =
  Expor.register
    (lift_set_config
       (fun () ->
         Sklstream.map
           (fun (fc : Pgen.flagged_choice array) ->
             let choices = Array.map fst fc in
             let pool_sz = Options.get_pool_size () in
             let next = Pgen.next_vchoice pool_sz choices in
             (Pgen.get_vuniq next, next))))
;;

let next_perms_simpl =
  Expor.register
    (lift_set_config
       (fun () ->
         Sklstream.map
           (fun (fc : Pgen.flagged_choice array) ->
             let choices = Array.map fst fc in
             let pool_sz = Options.get_pool_size () in
             Pgen.next_vchoice pool_sz choices)))
;;

let make_prime =
  Expor.register
    (lift_set_config (fun () -> Sklstream.map (Cbits.make_prime ())))
;;

let snd_skl = Expor.register (lift_set_config (fun () -> Sklstream.map snd));;

let check_perms =
  Expor.register
    (lift_set_config
       (fun () ->
         let local_pool =
           Pool.create
             (Options.get_pool_directory ())
             (Options.get_temporary_directory ())
             (Options.get_pool_size ()) in
         let prime_checker = Cbits.check_product local_pool in
         Sklstream.map
           (fun (choice : Pgen.choice) ->
             let pool_data = List.map (Pool.find local_pool) choice in
             let mods =
               let transpose arr =
                 let row i = Array.map (fun a -> a.(i)) arr in
                 Array.init (Array.length arr.(0)) (fun i -> row i) in
               transpose (Array.of_list (List.map snd pool_data))
             in
             let lazy_mods =
               Array.mapi
                 (fun i m ->
                   lazy
                     (succ
                        (Array.fold_left
                           (fun out mi -> (mi * out) mod m)
                           2 mods.(i)) mod m))
                 Cbits.small_primes in
             let rec exists f a i =
               if i = Array.length a then false else
               f a.(i) || exists f a (i + 1) in
             let pr_info big =
               let rec pr_choice ch = function
               | [] -> ()
               | x :: l -> Printf.fprintf ch "%d %a" x pr_choice l in
               Printf.eprintf "Choice: %a%,%s\n%!"
                 pr_choice choice (if big then "(b)" else "") in
             if exists (fun x -> Lazy.force x = 0) lazy_mods 0
             then begin pr_info false; (choice, false) end
             else begin pr_info true; (choice, prime_checker choice) end)))
;;

let prime_generation_skeleton conf =
  let inst f = atom (conf, f) in
  let nw = Options.get_number_of_workers () in
  loop
    (inst stop_cond,
     inst next_perms |||
     (farm_vector (inst make_prime, nw) *** id ()) |||
     inst snd_skl |||
     Skl_extra.rails_same (inst check_perms, nw))
;;

let make_pool_skl conf =
  let inst f = atom (conf, f) in
  let nw = Options.get_number_of_workers () in
  farm_vector (inst make_prime, nw)
;;

let explore_combinations conf =
  let inst f = atom (conf, f) in
  let nw = Options.get_number_of_workers () in
  loop
    (inst stop_cond,
     inst next_perms_simpl ||| Skl_extra.rails_same (inst check_perms, nw))
;;

type algo =
| Pool_first
| Pool_on_demand
;;

let main algo { Parfun.parfun = parfun; } =
  Ucommand_option.parse ();
  let init =
    let n_fact = Options.get_number_of_factors () in
    let pool_sz = Options.get_pool_size () in
    let wn = Options.get_number_of_workers () in
    let initv =
      match Options.get_recover_data () with
      | None -> Pgen.init_vchoice n_fact pool_sz wn
      | Some ch ->
        prerr_endline "** Recover mode enabled.";
        Pgen.recover_vchoice ch pool_sz wn
    in
    Sklstream.singleton (Array.map (fun x -> (x, false)) initv)
  in
  let local_pool =
    Pool.create
      (Options.get_pool_directory ())
      (Options.get_temporary_directory ())
      (Options.get_pool_size ())
  in
  let mk_prod = Cbits.make_product local_pool in
  let conf = Options.save_config () in
  let output =
    match algo with
    | Pool_on_demand ->
      let algo_skl = prime_generation_skeleton conf in
      Sklstream.to_list (parfun algo_skl init)
    | Pool_first ->
      let is =
        Sklstream.singleton
          (Array.of_list (Pgen.init_choice (Options.get_pool_size ()))) in
      Sklstream.iter ignore (parfun (make_pool_skl conf) is);
      Sklstream.to_list (parfun (explore_combinations conf) init)
  in
  if output = [] then failwith "try later !" else
  List.iter
    (Array.iter
       (fun (choice, b) ->
         if b then
           let prod = mk_prod choice in
           Printf.printf "found: %s\n\nHere are the factors of (p-1)/2:\n" prod;
           List.iter
             (fun i ->
               let (fact, _) = Pool.find local_pool i in
               Printf.printf " + %s\n" fact)
             choice;
           let rec display_choice chan = function
           | [] -> assert false
           | [ x ] -> Printf.fprintf chan "%d" x
           | x :: l ->
             Printf.fprintf chan "%d:%a" x display_choice l
           in
           Printf.printf
             "\nTo resume this computation, use option --recover %a\n"
             display_choice choice
         else
           ()))
    output
;;

Options.add_options ();;
Parfun.pardo (main Pool_on_demand);;
