(* $Id: fj.ml,v 1.22 2012/06/05 12:20:39 fclement Exp $ *)

open Fj_core;;
open Streams;;

type pid = int;;

type ('a, 'b) fjtree = ('a Streams.t, 'b Streams.t) cmp_tree;;

type 'a internal_out_channel = Pervasives.out_channel;;
type 'a internal_in_channel = Pervasives.in_channel;;

type 'a fj_out_channel = 'a option internal_out_channel;;
type 'a fj_in_channel = 'a option internal_in_channel;;

type ('a, 'b) cfjtree =
  | CFJ of ('a, 'b) cfjtree * ('a, 'b) cfjtree (** *)
  | CSeq of 'a fj_in_channel * 'b fj_out_channel (** *)
;;
(** A [('a, 'b) cfjtree] is the compiled form of a [('a, 'b) fjtree]:
  it is now a bunch of processes with each process ios connectors (the
  standard input and standard output channels) are stored into the
  data structure. A simple implementation can be to allocate pipes to
  connect the workers. *)

type ('a, 'b) fork_data =
  | Child of 'a fj_in_channel * 'b fj_out_channel (** *)
  | Parent of 'b fj_in_channel * 'a fj_out_channel * pid (** *)
;;
(** A ['a fork_data] represents the connectors involves in a fork
  system command: the [Child] summand for the child process, and the
  [Parent] summand for the parent process. *)

type 'a packet =
   | Data of 'a
   | End
   | Start
;;
(** A ['a packet] is an object travelling on wires, it is used to
  serialize streams. *)


(* SYSTEM SWISS ARMY KNIFE *)

let cpipe () =
  let id, od = Unix.pipe () in
  Unix.in_channel_of_descr id, Unix.out_channel_of_descr od
;;

let ufork () =
  let ic1, oc1 = cpipe () in
  let ic2, oc2 = cpipe () in
  match Unix.fork () with
  | 0 -> Child (ic1, oc2)
  | pid -> Parent (ic2, oc1, pid)
;;

let killall pids =
  List.iter (fun x -> Unix.kill x 15; ignore (Unix.waitpid [] x)) pids;
;;

let kill_on_term pids _x =
  killall pids;
  ignore (Sys.signal 15 Sys.Signal_default); Unix.sleep 1;
  Unix.kill (Unix.getpid ()) 15
;;


(* COMMUNICATION TOOLS *)

let rec close_tree t =
  match t with
  | CFJ (t1, t2) ->
    close_tree t1; close_tree t2
  | CSeq (ic, oc) ->
    close_in ic; close_out oc
;;

let send_in_tree tree s =
  let rec round_send tree s =
    match tree with
    | CFJ (t1, t2) ->
      let stream1, stream2 = fork s in
      let os1 = round_send t1 stream1 in
      let os2 = round_send t2 stream2 in
      join (os1, os2)
    | CSeq (_, oc) -> stream_to_channel s oc in
  iter ignore (round_send tree s);
  close_tree tree
;;

let rec receive_in_tree tree =
  match tree with
  | CFJ (t1, t2) ->
    let r1 = receive_in_tree t1 in
    let r2 = receive_in_tree t2 in
    join (r1, r2)
  | CSeq (ic, _) ->
    of_channel ic
;;


(* SKELETONS CODE *)

(* nearly parfun *)
let xparfun ctree is pids =
  let send_data () =
    send_in_tree ctree is;
    exit 0
  and output_stream () =
    receive_in_tree ctree in
  match Unix.fork () with
  | 0 -> (* child *) send_data ()
  | pid -> (* parent *) output_stream (), pid :: pids
;;

let run_function ic f oc =
  to_channel (f (of_channel ic)) oc;
;;

let rec start_tree :
    'a 'b. ('a, 'b) fjtree -> pid list -> ('a, 'b) cfjtree * pid list =
  fun tree pids ->
  match tree with
  | Par dyn ->
    let t1, t2 = get_par dyn in
    let ct1, pids1 = start_tree t1 pids in
    let ct2, pids2 = start_tree t2 pids1 in
    CFJ (ct1, ct2), pids2
  | Pipe dyn -> begin
    let t1, t2 = get_pipe dyn in
    match ufork () with
    | Child (ic, oc) ->
      let ct1, pidl = start_tree t1 [] in
      let ct2, pidl = start_tree t2 pidl in
      let os1, pidl = xparfun ct1 (of_channel ic) pidl in
      let os, pidl = xparfun ct2 os1 pidl in
      ignore (Sys.signal 15 (Sys.Signal_handle (kill_on_term pidl)));
      to_channel os oc;
      exit 0
    | Parent (ic, oc, pid) ->
      CSeq (ic, oc), pid :: pids end
  | PPP dyn -> begin
    let pre, tree, post = get_ppp dyn in
    match ufork () with
    | Child (ic, oc) ->
      let ct, pidl = start_tree tree [] in
      let is = pre (of_channel ic) in
      let os, pidl = xparfun ct is pidl in
      let os = post os in
      ignore (Sys.signal 15 (Sys.Signal_handle (kill_on_term pidl)));
      to_channel os oc;
      exit 0
    | Parent (ic, oc, pid) ->
      CSeq (ic, oc), pid :: pids end
  | Atom dyn -> begin
    let f = get_atom dyn in
    match ufork () with
    | Child (ic, oc) ->
      run_function ic f oc; exit 0
    | Parent (ic, oc, pid) ->
      CSeq (ic, oc), pid :: pids end
;;

let seq f = Fj_core.atom (Streams.map f);;

let pipeline a b = Fj_core.pipe (a, b);;

let ( ||| ) = pipeline;;

let collect (x1, x2) = Fj_core.par (x1, x2);;

let pre_post_process (pre, skl, post) = Fj_core.ppp (pre, skl, post);;

let par_apply x1 x2 = x1,x2;;

let in_map collect (skl, n) =
  if n <= 0 then failwith "map: non positive number of workers." else
  let rec in_map n =
    if n = 1 then skl else
    let r = n mod 2
    and q = n / 2 in
    collect
      (par_apply
        (in_map (q + r))
        (in_map q)) in
  in_map n
;;

let map (skl, n) = in_map collect (skl, n);;

let map_vector (skl, n) =
  let fskl = map (skl, n) in
  pre_post_process (Streams.of_sarray, fskl, Streams.to_sarray)
;;

(* helpers *)

let parfun_pids = ref [];;

let unseq tree x = (* FIXME _pids *)
  let ctree, _pids = start_tree tree [] in
  let is = Streams.singleton x in
  send_in_tree ctree is;
  let os = receive_in_tree ctree in
  let y = Streams.head os in
  y
;;

let parfun tree is =
  let ctree, pids = start_tree tree !parfun_pids in
  let os, pids = xparfun ctree is pids in
  parfun_pids := pids;
  os
;;
(** To be correct, [parfun] must spawn a computational network at each
  invocation with an input stream argument. *)

let terminate () =
  killall !parfun_pids;
  parfun_pids := []
;;
